MODULE maximum_multipole
  INTEGER :: lmax=100
  DOUBLE PRECISION :: mQ=0.05d0 ! m/q where q is comoving momentum of neutrinos.
                               ! Neutrino mass becomes important when z ~ zeq*mQ
  DOUBLE PRECISION :: kC=1d1 ! k in units of keq=0.01 h Mpc^-1.
                               ! The matter-radiation equality occurs when x=k*eta ~ kC.
  DOUBLE PRECISION :: z,xini=1d-2 ! initial k*eta
END MODULE maximum_multipole
PROGRAM Massive
  USE maximum_multipole
!
! This program solves differential equations of Boltzmann hierarchy 
! for massive neutrino perturbations.
!
! The differntial equation is solved by a publicly available routine 
! called "dverk.f".  For more information on dverk.f, see
! http://www.netlib.org/ode/dverk.f
!
! The output files contain:
! 1st column: x=k*tau (where tau is the conformal time)
! 2nd column: approximate solution with lmax=100
! 3rd column: exact solution (Shoji&Komatsu 2010)
!
! E.Komatsu, October 15, 2009
!
  IMPLICIT none
  DOUBLE PRECISION, allocatable :: y(:),work(:,:)
  DOUBLE PRECISION :: x,xstart,c(24),tol=1d-8,analytical(3),rombint
  DOUBLE PRECISION :: j0,j1,j2,j3,j4
  DOUBLE PRECISION :: a,term,func,source0,source1,source2,eqi
  INTEGER :: i,ind,n
  EXTERNAL fderiv,func,source0,source1,source2
  n=lmax+1 ! number of equations to solve
  allocate(y(n),work(n,9))
  open(1,file='solution_psi0.txt')
  open(2,file='solution_psi1.txt')
  open(3,file='solution_psi2.txt')
  ind=1
  do i=1,300
     x = xini*10d0**((dble(i)-1d0)*1d-2) ! x = k*eta
     if (i.eq.1) then 
        xstart = x
        a = xstart/kC+(xstart/kC/2d0)**2d0
        eqi = (1.d0+a**2.d0*mQ**2.d0)**0.5d0 ! eps(x_i)/q
        y(1) = 0.5d0  ! set initial condition for Psi0=0.5
        y(2) = -xini*eqi/6d0 ! set initial condition for Psi1=-x_i/6
        y(3) = -xini**2d0/30d0 ! set initial condition for Psi2=-x_i^2/30*eps(x_i)/q
        y(4:lmax) = 0d0 ! set initial conditions for Psi(l) for l>=3
     else
        call dverk(n,fderiv,xstart,y,x,tol,ind,c,n,work)
        xstart = x
        ! define spherical bessel functions
        !!$ z=x-xini
	z=rombint(func,dlog(xini),dlog(x),1d-7)
	if(z.gt.1d-2)then
           j0=dsin(z)/z
           j1=dsin(z)/z**2d0-dcos(z)/z
           j2=(3d0/z**3d0-1d0/z)*dsin(z)-3d0/z**2d0*dcos(z)
	   j3=(15d0/z**4d0-6d0/z**2)*dsin(z)+(1d0/z-15d0/z**3d0)*dcos(z)
	   j4=(1d0/z-45d0/z**3d0+105d0/z**5d0)*dsin(z)+(10d0/z**2d0-105d0/z**4d0)*dcos(z)
        else
           j0=1d0-z**2d0/6d0+z**4d0/120d0
           j1=z/3d0-z**3d0/30d0
           j2=z**2d0/15d0-z**4d0/210d0
	   j3=z**3d0/105d0
           j4=z**4d0/945d0
        endif
        ! analytical solutions (psi=1)
        a=x/kC+(x/kC/2d0)**2d0    ! kC = k/C and C=0.01 h Mpc^-1
        term=mQ**2d0*rombint(source0,xini,x,1d-7)
        analytical(1)= 1.d0+(a*mQ)**2.d0+(0.5d0-eqi**2.d0)*j0 + 0.5d0*eqi*xini*j1 - xini**2d0/6d0*j2 - term
        term=mQ**2d0*rombint(source1,xini,x,1d-7)
        analytical(2)= -(xini/6d0)*eqi*(j0-2d0*j2) + (xini**2d0/15d0+0.5d0-eqi**2.d0)*j1 - (xini**2d0/10d0)*j3 - term
        term=mQ**2d0*rombint(source2,xini,x,1d-7)
        analytical(3)= (0.5d0-eqi**2.d0)*j2-eqi*xini*j1/5.d0 &
             -(xini**2d0/30)*(j0-1.d1/7.d0*j2+18d0/7d0*j4) + 3*eqi*xini/10d0*j3 -term
        ! output the results
        write(1,'(3E20.8)')x,y(1),analytical(1)
        write(2,'(3E20.8)')x,y(2),analytical(2)
        write(3,'(3E20.8)')x,y(3),analytical(3)
     endif
  enddo
  close(1)
  deallocate(y,work)
  stop
END PROGRAM Massive
!-----------------------------------------------------------------------------
SUBROUTINE fderiv(n,x,y,yprime)
  USE maximum_multipole
!
! Massive neutrinos:
!
! Psi_0' = -(Q/E)Psi_1
! Psi_1' = (Q/3E)(Psi_0-2*Psi_2) - psi*(E/3Q)
! Psi_l' = (Q/E)/(2l+1)[l*Psi_{l-1}-(l+1)*Psi_{l+1}] for l>=2
! 
! where psi=1, derivatives are with respect to x=k*eta, and
!
! y(1)=Psi_0
! y(2)=Psi_1
! y(3)=Psi_2
! etc.
!
! Here, (E/Q)^2=1+(m*a/Q)^2, a=(C/k)x+[(C/k)x]^2/4, x=k*eta,
! and C=0.01 h Mpc^-1. 
!
  IMPLICIT none 
  INTEGER, intent(IN) :: n ! # of derivatives
  DOUBLE PRECISION :: x,y(n),yprime(n)
  DOUBLE PRECISION :: ell,EQ,a
  INTEGER :: l,j
  a=x/kC+(x/kC/2d0)**2d0    ! kC = k/C and C=0.01 h Mpc^-1
  EQ=dsqrt(1d0+(mQ*a)**2d0) ! mQ = m/Q
  yprime(1)= -y(2)/EQ 
  yprime(2)= ( y(1)-2d0*y(3) )/3d0/EQ - EQ/3d0
  do l=3,lmax+1
     ell=dble(l-1)
     yprime(l)= ( ell*y(l-1)-(ell+1d0)*y(l+1) )/(2d0*ell+1d0)/EQ
  enddo
  return
END SUBROUTINE fderiv
!-----------------------------------------------------------------------------
DOUBLE PRECISION FUNCTION func(lnx)
  USE maximum_multipole
! This function is the integrand of: 
! int dx(Q/E) = int dx/sqrt( 1+m^2*a^2/Q^2 )
! where a=(C*eta)+(C*eta)^2/4=(C/k)*x+[(C/k)x]^2/4
  IMPLICIT none
  DOUBLE PRECISION, intent(IN) :: lnx
  DOUBLE PRECISION :: a,x
  x=dexp(lnx)
  a=x/kC+(x/kC/2d0)**2d0    ! kC = k/C and C=0.01 h Mpc^-1
  func=x/dsqrt(1d0+(mQ*a)**2d0) ! mQ = m/Q
  return
END FUNCTION func
!-----------------------------------------------------------------------------
DOUBLE PRECISION FUNCTION source0(x)
  USE maximum_multipole
! This function is the integrand of: 
! (2C/k)*int dx' a(x')*[1+(C/2k)*x']*j_0[z(x)-z(x')]
! where
! z = int dx(Q/E) = int dx/sqrt( 1+m^2*a^2/Q^2 )
  IMPLICIT none
  DOUBLE PRECISION, intent(IN) :: x
  DOUBLE PRECISION :: a,j0,func,zp,zz
  external func
  a=x/kC+(x/kC/2d0)**2d0 ! kC = k/C and C=0.01 h Mpc^-1
  if(xini.eq.x)then
     zp=0d0
  else
     call qsimp(func,dlog(xini),dlog(x),zp)
  endif
  zz=z-zp
  if(abs(zz).gt.1d-2)then
     j0=dsin(zz)/zz
  else
     j0=1d0-zz**2d0/6d0+zz**4d0/120d0
  endif
  source0=(2d0/kC)*a*(1d0+0.5d0*x/kC)*j0
  return
END FUNCTION source0
!-----------------------------------------------------------------------------
DOUBLE PRECISION FUNCTION source1(x)
  USE maximum_multipole
! This function is the integrand of: 
! (2C/k)*int dx' a(x')*[1+(C/2k)*x']*x'*j_1[z(x)-z(x')]
! where
! z = int dx(Q/E) = int dx/sqrt( 1+m^2*a^2/Q^2 )
  IMPLICIT none
  DOUBLE PRECISION, intent(IN) :: x
  DOUBLE PRECISION :: a,j1,func,zp,zz
  external func
  a=x/kC+(x/kC/2d0)**2d0 ! kC = k/C and C=0.01 h Mpc^-1
  if(xini.eq.x)then
     zp=0d0
  else
     call qsimp(func,dlog(xini),dlog(x),zp)
  endif
  zz=z-zp
  if(abs(zz).gt.1d-2)then
     j1=dsin(zz)/zz**2d0-dcos(zz)/zz
  else
     j1=zz/3d0-zz**3d0/30d0
  endif
  source1=(2d0/kC)*a*(1d0+0.5d0*x/kC)*j1
  return
END FUNCTION source1
!-----------------------------------------------------------------------------
DOUBLE PRECISION FUNCTION source2(x)
  USE maximum_multipole
! This function is the integrand of: 
! (2C/k)*int dx' a(x')*[1+(C/2k)*x']*x'*j_2[z(x)-z(x')]
! where
! z = int dx(Q/E) = int dx/sqrt( 1+m^2*a^2/Q^2 )
  IMPLICIT none
  DOUBLE PRECISION, intent(IN) :: x
  DOUBLE PRECISION :: a,j2,func,zp,zz
  external func
  a=x/kC+(x/kC/2d0)**2d0 ! kC = k/C and C=0.01 h Mpc^-1
  if(xini.eq.x)then
     zp=0d0
  else
     call qsimp(func,dlog(xini),dlog(x),zp)
  endif
  zz=z-zp
  if(abs(zz).gt.1d-2)then
     j2=(3d0/zz**3d0-1d0/zz)*dsin(zz)-3d0/zz**2d0*dcos(zz)
  else
     j2=zz**2d0/15d0-zz**4d0/210d0
  endif
  source2=(2d0/kC)*a*(1d0+0.5d0*x/kC)*j2
  return
END FUNCTION source2
